<?php

declare(strict_types=1);

namespace Drupal\sync_client_example\Plugin\SyncClient;

use Drupal\Core\Form\FormStateInterface;
use Drupal\sync_clients\SyncClient\PagedImportSyncClientBase;
use GuzzleHttp\RequestOptions;
use Psr\Http\Message\ResponseInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Example Sync Client corresponding with SyncClientSimpleImport.
 *
 * @SyncClient(
 *   id = "sync_client_simple_paged_import",
 *   label = @Translation("Simple Paged Import"),
 *   priority = 1,
 *   queue_id = "sync_client_simple_paged_import",
 *   job_type_id = "sync_client_simple_paged_import"
 * )
 */
class SyncClientSimplePagedImport extends PagedImportSyncClientBase implements SyncClientSimplePagedImportInterface {

  /**
   * {@inheritdoc}
   */
  public function addPluginFields(array &$form, FormStateInterface $form_state): void {
    parent::addPluginFields($form, $form_state);

    // @todo Add any other value you need to be configurable.
    $form['url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('REST Endpoint URL'),
      '#description' => $this->t('The full URL of the endpoint.'),
      '#default_value' => $this->config->get('url') ?? '',
      '#required' => TRUE,
    ];

    $form['items_page'] = [
      '#type' => 'number',
      '#title' => $this->t('Page size to request from the API'),
      '#description' => $this->t('The number amount of items we want the API to return per page, when running a full sync.'),
      '#default_value' => $this->config->get('items_page') ?? 25,
      '#min' => 1,
      '#max' => 100,
      '#step' => 1,
      '#weight' => 9,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function get(int $page = 0, array $filter_parameters = []): ResponseInterface {
    return $this->request(Request::METHOD_GET, $this->configuration['url'], $this->buildOptions($page, 0, $filter_parameters));
  }

  /**
   * Build the options for the request.
   *
   * @param int $page
   *   The page to fetch. Defaults to 0 (first page)
   * @param int $size
   *   The page size to request. Defaults to 0, in which case the plugin config
   *   is used.
   * @param string[] $filter_parameters
   *   Query parameters for filtering.
   *
   * @return array<string, array<string, mixed>>
   *   The request options.
   */
  private function buildOptions(int $page = 0, int $size = 0, array $filter_parameters = []): array {

    if ($size < 1) {
      $size = $this->configuration['items_page'];
    }

    $options = [
      RequestOptions::HEADERS => [
        'Accept' => 'application/json',
        'Cache-Control' => 'no-cache',
        'Content-Type' => 'application/json',
      ],
      RequestOptions::QUERY => [
        'size' => $size,
        'page' => $page,
      ],
    ];

    if (count($filter_parameters) > 0) {
      $options[RequestOptions::QUERY] = array_merge($options[RequestOptions::QUERY], $filter_parameters);
    }

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  protected function doPing(): ResponseInterface {
    return $this->request(Request::METHOD_GET, $this->configuration['url'], $this->buildOptions(), TRUE);
  }

}
