<?php

declare(strict_types=1);

namespace Drupal\sync_clients\SyncClient;

use Drupal\Core\Form\FormStateInterface;

/**
 * Base implementation of Sync Clients with progress tracking for paged imports.
 */
abstract class PagedImportSyncClientBase extends SyncClientBase implements PagedImportSyncClientInterface {

  /**
   * {@inheritdoc}
   */
  public function addPluginFields(array &$form, FormStateInterface $form_state): void {
    parent::addPluginFields($form, $form_state);

    $form['temp_storage_expiration'] = [
      '#type' => 'number',
      '#title' => $this->t('Progress/cleanup storage expiration'),
      '#description' => $this->t('The duration the progress state (last fully processed API response) should be kept on error, in seconds.<br>This allows a rerun, after fixing the problem, to start from last successful point instead of the beginning. The state is always cleared after completion of a run.'),
      '#default_value' => $this->config->get('temp_storage_expiration') ?? 3600,
      '#min' => 0,
      '#max' => 82800,
      '#step' => 60,
      '#weight' => 10,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function deleteProgressData(): void {
    $this->keyValueExpirable->get($this->getPluginId())->delete("progress_data_{$this->getPluginId()}");
  }

  /**
   * {@inheritdoc}
   */
  public function getFetchProgress(): int {
    $data = $this->getFetchProgressData();
    return $data['processed_page'] ?? 0;
  }

  /**
   * {@inheritdoc}
   */
  public function setFetchProgressData(int $page = 0): void {
    $data = $this->getFetchProgressData();

    if (count($data) === 0) {
      $data['started'] = time();
    }
    $data['processed_page'] = $page;

    $this->keyValueExpirable->get($this->getPluginId())->setWithExpire("progress_data_{$this->getPluginId()}", $data, $this->configuration['temp_storage_expiration']);
  }

  /**
   * Get the fetch progress data of current/last fetch run.
   *
   * @return array<string, int>
   *   Array containing 'started' timestamp and 'processed_page' being the last
   *   fetch and processed page. Empty array if not set or expired.
   */
  protected function getFetchProgressData(): array {
    return $this->keyValueExpirable->get($this->getPluginId())->get("progress_data_{$this->getPluginId()}", []);
  }

}
